/**
 * \brief Example of how to read a file (list of files) using StFemtoEvent classes
 *
 * RunFemtoDstAnalyzer.C is an example of reading FemtoDst format.
 * One can use either FemtoDst file or a list of femtoDst files (inFile.lis or
 * inFile.list) as an input, and preform physics analysis
 *
 * \author Grigory Nigmatkulov
 * \date May 29, 2018
 */

// This is needed for calling standalone classes
#define _VANILLA_ROOT_

// C++ headers
#include <iostream>
#include <vector>
#include <map>

// ROOT headers
#include "TROOT.h"
#include "TFile.h"
#include "TChain.h"
#include "TVector2.h"
#include "TVector3.h"
#include "TTree.h"
#include "TSystem.h"
#include "TH1.h"
#include "TH2.h"
#include "TMath.h"
#include "TProfile2D.h"
#include "TStopwatch.h"

// FemtoDst headers
#include "/mnt/pool/rhic/4/parfenovpeter/STAR/StFemtoEvent/StFemtoDstReader.h"
#include "/mnt/pool/rhic/4/parfenovpeter/STAR/StFemtoEvent/StFemtoDst.h"
#include "/mnt/pool/rhic/4/parfenovpeter/STAR/StFemtoEvent/StFemtoEvent.h"
#include "/mnt/pool/rhic/4/parfenovpeter/STAR/StFemtoEvent/StFemtoTrack.h"

// Load libraries (for ROOT_VERSTION_CODE >= 393215)
#if ROOT_VERSION_CODE >= ROOT_VERSION(6, 0, 0)
R__LOAD_LIBRARY(/mnt/pool/rhic/4/parfenovpeter/STAR/build/libStFemtoDst.so)
#endif

#include "Constants.h"

// inFile - is a name of name.FemtoDst.root file or a name
//          of a name.lis(t) files, that contains a list of
//          name1.FemtoDst.root, name2.FemtoDst.root, ... files

//Used functions (see them below)
Bool_t isGoodEvent(StFemtoEvent *const &event);
Bool_t isGoodTrack(StFemtoTrack *const &track, Float_t _energy, TVector3 pVtx);
Bool_t isGoodTrackFlow(StFemtoTrack *const &track, Float_t _energy, TVector3 pVtx);
Double_t GetWeight(StFemtoTrack *const &track);
TVector2 CalcQvector(StFemtoTrack *const &track, Int_t _harm);
Double_t AngleShift(Double_t Psi, Double_t order);

//_________________
void FemtoDstAnalyzer_QA(const Char_t *inFile = "st_physics_12150008_raw_4030001.femtoDst.root",
                                         const Char_t *outFile = "./oShiftTest.root")
{

  std::cout << "Hi! Lets do some physics, Master!" << std::endl;
  TStopwatch timer;
  timer.Start();

#if ROOT_VERSION_CODE < ROOT_VERSION(6, 0, 0)
  gSystem->Load("../libStFemtoDst.so");
#endif

  StFemtoDstReader *femtoReader = new StFemtoDstReader(inFile);
  femtoReader->Init();

  // This is a way if you want to spead up IO
  std::cout << "Explicit read status for some branches" << std::endl;
  femtoReader->SetStatus("*", 0);
  femtoReader->SetStatus("Event", 1);
  femtoReader->SetStatus("Track", 1);
  std::cout << "Status has been set" << std::endl;

  std::cout << "Now I know what to read, Master!" << std::endl;

  if (!femtoReader->chain())
  {
    std::cout << "No chain has been found." << std::endl;
  }
  Long64_t eventsInTree = femtoReader->tree()->GetEntries();
  std::cout << "eventsInTree: " << eventsInTree << std::endl;
  Long64_t events2read = femtoReader->chain()->GetEntries();

  std::cout << "Number of events to read: " << events2read << std::endl;

  //Eventwise parameters
  //hHisto[0] - before event selection
  //hHisto[1] - after event selection
  TH2F *hVtxXY[2];
  TH1F *hVtxZ[2];
  TH1F *hVtxVpdZ[2];

  //Trackwise parameters
  //hHisto[0] - before track selection (but after event selection)
  //hHisto[1] - after track selection
  TH1F *hNhitsFit[2];
  TH1F *hNhitsPoss[2];
  TH1F *hNhitsRatio[2];
  TH1F *hPt[2];
  TH1F *hPtot[2];
  TH1F *hEta[2];
  TH1F *hDCA[2];

  //PID related parameters
  //hHisto[0] - before any PID (but after track selection)
  //hHisto[1] - after PID
  //second [i] is for particle species: 0-all, 1-pion, 2-kaon, 3-proton
  TH2F *hnSigM2_lowpt[2];
  TH2F *hnSigM2_highpt[2];
  TH2F *hdEdxQp[2][4];
  TH2F *hM2Qp[2][4];
  TH2F *hNsigQp[2][4];
  TH2F *hM2pt[2][4];
  TH2F *hNsigpt[2][4];

  //Initialization
  for (int i=0; i<2; i++)
  {
    //Eventwise
    hVtxXY[i]   = new TH2F(Form("hVtxXY%i",i),Form("vertex XY %i;Vtx_{X}, [cm];Vtx_{Y}, [cm]",i),200,-5.,5.,200,-5.,5.);
    hVtxZ[i]    = new TH1F(Form("hVtxZ%i",i) ,Form("vertex Z %i;Vtx_{Z}, [cm];N_{counts}",i)    ,200,-40.,40.);
    hVtxVpdZ[i] = new TH1F(Form("hVtxVpdZ%i",i) ,Form("VtxZ - VpdVz %i;Vtx_{Z}-vpdV_{Z}, [cm];N_{counts}",i),200,-5.,5.);
    //Trackwise
    hNhitsFit[i]= new TH1F(Form("hNhitsFit%i",i) ,Form("N_{hist}^{Fit} %i;N_{hist}^{Fit};N_{counts}",i),50,0.,50.);
    hNhitsPoss[i]= new TH1F(Form("hNhitsPoss%i",i) ,Form("N_{hist}^{Poss} %i;N_{hist}^{Poss};N_{counts}",i),50,0.,50.);
    hNhitsRatio[i]= new TH1F(Form("hNhitsRatio%i",i) ,Form("N_{hist}^{Fit}/N_{hist}^{Poss} %i;N_{hist}^{Fit}/N_{hist}^{Poss};N_{counts}",i),50,0.,1.);
    hPt[i] = new TH1F(Form("hPt%i",i) ,Form("p_{T} %i;p_{T}, [GeV/c];N_{counts}",i),600,0.,6.);
    hPtot[i] = new TH1F(Form("hPtot%i",i) ,Form("p_{tot} %i;p, [GeV/c];N_{counts}",i),600,0.,12.);
    hEta[i] = new TH1F(Form("hEta%i",i) ,Form("#eta %i;#eta;N_{counts}",i),400,-2.,2.);
    hDCA[i] = new TH1F(Form("hDCA%i",i) ,Form("|DCA| %i;dca, [cm];N_{counts}",i),200,0.,5.);
    //PID
    hnSigM2_lowpt[i] = new TH2F(Form("hnSigM2_lowpt%i",i) ,Form("n#sigma_{#pi} vs M^{2} for 0.5 < p_{T} < 0.7 GeV/c %i;n#sigma_{#pi};m^{2} [GeV/c^{2}]^{2}",i),600,-16.,16.,600,-0.8,1.8);
    hnSigM2_highpt[i] = new TH2F(Form("hnSigM2_highpt%i",i) ,Form("n#sigma_{#pi} vs M^{2} for 2.2 < p_{T} < 2.4 GeV/c %i;n#sigma_{#pi};m^{2} [GeV/c^{2}]^{2}",i),650,-7.,6.,600,-0.8,1.8);
    for (int iPID=0; iPID<4; iPID++)
    {
      hdEdxQp[i][iPID] = new TH2F(Form("hdEdxQp%i%i",i,iPID) ,Form("dEdx vs Q*p_{tot} %i %i;Q*p, [GeV/c];dEdx, [a.u.]",i,iPID),1000,-5.,5.,1000,0.,2e-5);
      hM2Qp[i][iPID] = new TH2F(Form("hM2Qp%i%i",i,iPID) ,Form("m^{2} vs Q*p_{tot} %i %i;Q*p, [GeV/c];m^{2}, [GeV/c^{2}]^{2}",i,iPID),1000,-5.,5.,1000,-0.8,1.8);
      hNsigQp[i][iPID] = new TH2F(Form("hNsigQp%i%i",i,iPID) ,Form("n#sigma_{#pi} vs Q*p_{tot} %i %i;Q*p, [GeV/c];n#sigma_{#pi}",i,iPID),1000,-5.,5.,600,-3.,3.);
      hM2pt[i][iPID] = new TH2F(Form("hM2pt%i%i",i,iPID) ,Form("m^{2} vs p_{T} %i %i;p_{T}, [GeV/c];m^{2}, [GeV/c^{2}]^{2}",i,iPID),1000,0.,5.,1000,-0.8,1.8);
      hNsigpt[i][iPID] = new TH2F(Form("hNsigpt%i%i",i,iPID) ,Form("n#sigma_{#pi} vs p_{T} %i %i;p_{T}, [GeV/c];n#sigma_{#pi}",i,iPID),1000,0.,5.,600,-3.,3.);
    }
  }

  // Loop over events
  for (Long64_t iEvent = 0; iEvent < events2read; iEvent++)
  {

    if ((iEvent+1) % 1000 == 0){
      std::cout << "Working on event #[" << (iEvent + 1)
                << "/" << events2read << "]" << std::endl;
    }

    Bool_t readEvent = femtoReader->readFemtoEvent(iEvent);
    if (!readEvent)
    {
      std::cout << "Something went wrong, Master! Nothing to analyze..." << std::endl;
      break;
    }

    // Retrieve femtoDst
    StFemtoDst *dst = femtoReader->femtoDst();

    // Retrieve event information
    StFemtoEvent *event = dst->event();
    if (!event)
    {
      std::cout << "Something went wrong, Master! Event is hiding from me..." << std::endl;
      break;
    }

    TVector3 pVtx = event->primaryVertex();

    hVtxXY[0]->Fill(pVtx.X(),pVtx.Y());
    hVtxZ[0]->Fill(pVtx.Z());
    hVtxVpdZ[0]->Fill(pVtx.Z()-event->vpdVz());

    // Event selection
    if ( !isGoodEvent(event) ) continue;

    hVtxXY[1]->Fill(pVtx.X(),pVtx.Y());
    hVtxZ[1]->Fill(pVtx.Z());
    hVtxVpdZ[1]->Fill(pVtx.Z()-event->vpdVz());

    // Track analysis
    Int_t nTracks = dst->numberOfTracks();
    

    // Track loop
    for (Int_t iTrk = 0; iTrk < nTracks; iTrk++)
    {

      // Retrieve i-th femto track
      StFemtoTrack *femtoTrack = dst->track(iTrk);

      if (!femtoTrack)
        continue;

      hNhitsFit[0] -> Fill(femtoTrack->nHitsFit());
      hNhitsPoss[0] -> Fill(femtoTrack->nHitsPoss());
      if (femtoTrack->nHitsPoss() != 0){
        hNhitsRatio[0] -> Fill((Double_t) (femtoTrack->nHitsFit()/femtoTrack->nHitsPoss()));
      }
      hPt[0] -> Fill(femtoTrack->pt());
      hPtot[0] -> Fill(femtoTrack->ptot());
      hEta[0] -> Fill(femtoTrack->eta());
      hDCA[0] -> Fill(femtoTrack->gDCA(pVtx).Mag());

      // Must be a primary track
      if (!femtoTrack->isPrimary())
        continue;

   
      //Track selection
      if (!isGoodTrackFlow(femtoTrack, energy, pVtx)) continue;

      hNhitsFit[1] -> Fill(femtoTrack->nHitsFit());
      hNhitsPoss[1] -> Fill(femtoTrack->nHitsPoss());
      if (femtoTrack->nHitsPoss() != 0){
        hNhitsRatio[1] -> Fill((Double_t) (femtoTrack->nHitsFit()/femtoTrack->nHitsPoss()));
      }
      hPt[1] -> Fill(femtoTrack->pt());
      hPtot[1] -> Fill(femtoTrack->ptot());
      hEta[1] -> Fill(femtoTrack->eta());
      hDCA[1] -> Fill(femtoTrack->gDCA(pVtx).Mag());

      // Check if track has TOF signal
      if (femtoTrack->isTofTrack())
      {
        if (femtoTrack->gDCA(pVtx).Mag() >= 1.) continue;

        if (femtoTrack->pt() > 0.5 && femtoTrack->pt() < 0.7)
        {
          hnSigM2_lowpt[0]->Fill(femtoTrack->nSigmaPion(),femtoTrack->massSqr());
        }
        if (femtoTrack->pt() > 2.2 && femtoTrack->pt() < 2.4)
        {
          hnSigM2_highpt[0]->Fill(femtoTrack->nSigmaPion(),femtoTrack->massSqr());
        }

        hdEdxQp[0][0] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->dEdx());
        hM2Qp[0][0] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->massSqr());
        hNsigQp[0][0] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->nSigmaPion());
        hM2pt[0][0] -> Fill(femtoTrack->pt(),femtoTrack->massSqr());
        hNsigpt[0][0] -> Fill(femtoTrack->pt(),femtoTrack->nSigmaPion());

        if (femtoTrack->nSigmaPion() < 2.5)
        {
          hdEdxQp[0][1] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->dEdx());
          hM2Qp[0][1] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->massSqr());
          hNsigQp[0][1] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->nSigmaPion());
          hM2pt[0][1] -> Fill(femtoTrack->pt(),femtoTrack->massSqr());
          hNsigpt[0][1] -> Fill(femtoTrack->pt(),femtoTrack->nSigmaPion());
        }

        if (femtoTrack->nSigmaKaon() < 2.5)
        {
          hdEdxQp[0][2] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->dEdx());
          hM2Qp[0][2] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->massSqr());
          hNsigQp[0][2] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->nSigmaPion());
          hM2pt[0][2] -> Fill(femtoTrack->pt(),femtoTrack->massSqr());
          hNsigpt[0][2] -> Fill(femtoTrack->pt(),femtoTrack->nSigmaPion());
        }

        if (femtoTrack->nSigmaProton() < 2.5)
        {
          hdEdxQp[0][3] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->dEdx());
          hM2Qp[0][3] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->massSqr());
          hNsigQp[0][3] -> Fill(femtoTrack->charge()*femtoTrack->ptot(),femtoTrack->nSigmaPion());
          hM2pt[0][3] -> Fill(femtoTrack->pt(),femtoTrack->massSqr());
          hNsigpt[0][3] -> Fill(femtoTrack->pt(),femtoTrack->nSigmaPion());
        }

      } //if( isTofTrack() )

    } //for(Int_t iTrk=0; iTrk<nTracks; iTrk++)

  } //for(Long64_t iEvent=0; iEvent<events2read; iEvent++)

  femtoReader->Finish();

  TFile *output = new TFile(outFile,"recreate");

  for (int i=0; i<2; i++)
  {
    //Eventwise
    hVtxXY[i]   -> Write();
    hVtxZ[i]    -> Write();
    hVtxVpdZ[i] -> Write();
    //Trackwise
    hNhitsFit[i]-> Write();
    hNhitsPoss[i]-> Write();
    hNhitsRatio[i]-> Write();
    hPt[i] -> Write();
    hPtot[i] -> Write();
    hEta[i] -> Write();
    hDCA[i] -> Write();
    //PID
    hnSigM2_lowpt[i] -> Write();
    hnSigM2_highpt[i] -> Write();
    for (int iPID=0; iPID<4; iPID++)
    {
      hdEdxQp[i][iPID] -> Write();
      hM2Qp[i][iPID] -> Write();
      hNsigQp[i][iPID] -> Write();
      hM2pt[i][iPID] -> Write();
      hNsigpt[i][iPID] -> Write();
    }
  }

  output->Close();

  std::cout << "I'm done with analysis. We'll have a Nobel Prize, Master!"
            << std::endl;
  timer.Stop();
  timer.Print();
}

Bool_t isGoodEvent(StFemtoEvent *const &event)
{
  if (!event) return false;
  if (event == nullptr) return false;

  if (event->primaryVertex().Perp() > cutVtxR) return false;
  if (TMath::Abs(event->primaryVertex().Z()) > cutVtxZEnergy.at(energy)) return false;

  if ((energy == 200.) && TMath::Abs(event->primaryVertex().Z() - event->vpdVz()) > cutVpdVz) return false;

  return true;
}

Bool_t isGoodTrack(StFemtoTrack *const &track, Float_t _energy, TVector3 pVtx)
{
  if (!track) return false;
  // if (!track->isPrimary()) return false;
  if (track->nHitsFit() < cutNhits) return false;
  if (track->nHitsPoss() <= cutNhitsPoss) return false;
  if ((Double_t) track->nHitsFit()/track->nHitsPoss() < cutNhitsRatio) return false;
  if (TMath::Abs(track->eta()) >= cutEta) return false;
  
  if (track->pt() <= cutPtMin.at(_energy)) return false;
  if (track->pt() > cutPtMax) return false;
  if (track->ptot() > cutPMax) return false;
  if (track->gDCA(pVtx).Mag() >= cutDCA.at(_energy)) return false;
  return true;
}

Bool_t isGoodTrackFlow(StFemtoTrack *const &track, Float_t _energy, TVector3 pVtx)
{
  if (!track) return false;
  // if (!track->isPrimary()) return false;
  if (track->nHitsFit() < cutNhits) return false;
  if (track->nHitsPoss() <= cutNhitsPoss) return false;
  if ((Double_t) track->nHitsFit()/track->nHitsPoss() < cutNhitsRatio) return false;
  if (TMath::Abs(track->eta()) >= cutEta) return false;
  
  if (track->pt() <= cutPtMin.at(_energy)) return false;
  //if (track->pt() > cutPtMax) return false;
  if (track->ptot() > cutPMax) return false;
  if (track->gDCA(pVtx).Mag() >= cutDCA.at(_energy)) return false;
  return true;
}

Double_t GetWeight(StFemtoTrack *const &track)
{
  Double_t weight;
  if (track->pt() <= cutPtWeightEP)
  {
    weight = track->pt();
  }
  else
  {
    weight = cutPtWeightEP;
  }
  return weight;
}

TVector2 CalcQvector(StFemtoTrack *const &track, Int_t _harm)
{
  TVector2 qv(0.,0.);

  qv.Set(TMath::Cos(_harm*track->phi()),TMath::Sin(_harm*track->phi()));

  return qv;
}

Double_t AngleShift(Double_t Psi, Double_t order)
{
  Double_t PsiCorr = Psi;
  if (PsiCorr > TMath::Pi()/order)
  {
    PsiCorr = Psi - 2.*TMath::Pi()/order;
  }
  if (PsiCorr < -1.*TMath::Pi()/order)
  {
    PsiCorr = Psi + 2.*TMath::Pi()/order;
  }
  return PsiCorr;
}
